import numpy as np


def ecm(data):
    """Calculates the correlation between some pixel and the same pixel at time t+tau
    Make sure that the data that is fed to the ecm method is the time data at some pixel value.

    Parameters
    ----------
    data: array_like
        pixel values for some x,y position over a time series
    """
    # fast method uses a FFT and is a process which is O(n) = n log(n)
    data_length = len(data)
    data_fft = np.fft.fft(data)
    time_correlation = np.fft.ifft(data_fft * np.conjugate(data_fft)).real
    #print(len(time_correlation))
    # normalizing the data
    norm_correlation = time_correlation/data_length
    norm_correlation = norm_correlation/np.mean(data)**2
    return norm_correlation



import numpy as np
from empyer.misc.image import bin_2d


def angular_correlation(r_theta_img, mask=None, binning=1, cut_off=0, normalize=True):
    """A program that takes a 2d image and then preforms an angular correlation on the image.

    Parameters
    ----------
    r_theta_img: array_like
        The index of the axes
    mask: boolean array
        The name of the axis
    binning : int
        binning factor
    cut_off : int
        The cut off in pixels to
    normalize: bool
        Subtract <I(\theta)>^2 and divide by <I(\theta)>^2
    """
    image = r_theta_img
    if mask is not None:
        image[mask] = 0
        if cut_off is not 0:
            mask = mask[cut_off:len(mask), :]

    if cut_off is not 0:
        image = image[cut_off:len(image), :]

    if binning is not 1:
        image = bin_2d(image, binning)
        if mask is not None:
            mask = bin_2d(mask, binning) != 0



    # fast method uses a FFT and is a process which is O(n) = n log(n)
    I_fft = np.fft.fft(image, axis=1)
    a = np.fft.ifft(I_fft * np.conjugate(I_fft), axis=1).real
    # this is to determine how many of the variables were non zero... This is really dumb.  but...
    # it works and I should stop trying to fix it (wreak it)
    if mask is not None:
        mask_boolean = ~mask  # inverting the boolean mask
        mask_fft = np.fft.fft(mask_boolean, axis=1)
        number_unmasked = np.fft.ifft(mask_fft*np.conjugate(mask_fft), axis=1).real
        number_unmasked[number_unmasked == 0] = 1  # get rid of divide by zero error for completely masked rows
        a = np.multiply(np.divide(a, number_unmasked), 720)

    if normalize:
        a_prime = np.zeros(np.shape(a))
        for i, row in enumerate(a):
            row_mean = np.mean(row)
            if row_mean == 0:
                normalized_row = np.divide(np.subtract(row, row_mean), 1)
            else:
                normalized_row = np.divide(np.subtract(row, row_mean), row_mean)
                a_prime[i, :] = normalized_row
            a = a_prime
    return a


def power_spectrum(correlation, method="FFT"):
    """
    :param correlation: Taking the FFT of the angular correlation to find the symmetry present
    :param method: Right now this doesn't actually do anything but I want to add in other methods.
    :return: pow_spectrum: The resulting power spectrum from the angular correlation.  Gives indexes 0-180.
    """

    if method is "FFT":
        print(np.shape(correlation))
        pow_spectrum = np.fft.fft(correlation, axis=1).real
        pow_spectrum = np.power(pow_spectrum, 2)
    return pow_spectrum


def get_S_Q(r_theta_img,plot=False):

    # TODO: add in what the physical r is for the function...

    S_Q = [np.nansum(r) for r in r_theta_img]
    if plot:
        plt.plot(S_Q)
        plt.show()
    return S_Q





import numpy as np
from scipy.interpolate import RectBivariateSpline

from empyer.misc.image import ellipsoid_list_to_cartesian, polar_list_to_cartesian, create_grid


def convert(img, mask=None, center=None, angle=None, foci=None, radius=None, phase_width=720):
    #  This function someday will be faster hopefully...
    """
    :param img: a n by 2-d array for the image to convert to polar coordinates
    :param mask: 2-d boolean array which excludes certain points
    :param center: x,y coordinates for the center of the image
    :param angle: angle of rotation if the sample is elliptical
    :param foci: The lengths of the foci of the ellipse
    :param phase_width: the number of "pixels" in the polar image along the x direction
    :param radius: the number of "pixels" in the polar image along the y direction
    :param plot: Plot the image after converting it...
    :return: polar_img r vs. theta
    """
    img_shape = np.shape(img)
    if center is None:
        center = np.true_divide(img_shape[-2:], 2)
    center = [center[1], center[0]]  # converting to y,x or array coordinates
    if radius is None:
        radius = int(min(img_shape[-2:]) - max(np.abs(np.subtract(img_shape[-2:], center)))-5)
        if foci is not None:
            radius = int(radius/(max(foci)/min(foci)))
    # setting up grids for faster interpolation
    r_inital = 1
    r_final = int(radius + r_inital)

    initial_y, initial_x = range(1, img_shape[-2]+1), range(1, img_shape[-1]+1)
    # for perfectly circular conversions
    if angle is None and foci is None:
        final_theta, final_r = create_grid(np.linspace(-1 * np.pi, np.pi, phase_width), np.arange(r_inital, r_final, 1))
        final_x, final_y = polar_list_to_cartesian(final_r, final_theta, center)
    # for elliptical conversions
    else:
        angle = (np.pi/2)-angle  # converting to y,x or array coordinates
        final_the = np.linspace(-1*np.pi, np.pi, num=phase_width)
        final_rad = np.arange(r_inital, r_final, 1)
        final_x, final_y = ellipsoid_list_to_cartesian(final_rad,
                                                       final_the,
                                                       center,
                                                       major=foci[0],
                                                       minor=foci[1],
                                                       angle=angle,
                                                       even_spaced=True)
    inten = img

    # setting masked values to negative values. Anything interpolated from masked values becomes negative
    if mask is not None:
        inten[mask] = -999999
    # For higher than 2 dimensional arrays.  Speeds up computations a little but requires more memory
    if len(img_shape) > 2:
        inten = np.reshape(inten, (-1, *img_shape[-2:]))
        polar_img = np.zeros((inten.shape[0],)+(r_final - r_inital, phase_width))
        for i, img in enumerate(inten):
            spline = RectBivariateSpline(initial_x, initial_y, img, kx=1, ky=1)  # bi-linear spline
            polar = np.array(spline.ev(final_x, final_y))
            polar_img[i] = np.reshape(polar, [r_final - r_inital, phase_width])
        polar_img = np.reshape(polar_img, (*img_shape[:-2], radius, phase_width))

    else:
        spline = RectBivariateSpline(initial_x, initial_y, inten, kx=1, ky=1)  # bi-linear spline
        polar_img = np.array(spline.ev(final_x, final_y))
        polar_img = np.reshape(polar_img, (radius, phase_width))

    # outputting new mask
    if mask is not None:
        polar_mask = polar_img < 0
        polar_img[polar_img < 0] = 0

    else:
        polar_mask = None
    return polar_img






import numpy as np


def flatten_axis(array, axis):
    array_shape = np.shape(array)
    np.reshape(array, (*array_shape[:axis-1], -1, *array_shape[-2:]))


def square(array):
    return np.power(array, 2)


def bin_2d(image, binning_factor):
    """Binning a 2-dimensional image  by some factor

    Parameters
    ----------
    image : 2-d array

    binning_factor : int

    Returns
    ----------
    new_image : 2-d array
    """
    sh = np.shape(image)

    dim1_cut = sh[0]%binning_factor
    dim2_cut = sh[1]%binning_factor
    cut_image = image[dim1_cut:, dim2_cut:]  # in case the image is not a multiple of the binning factor

    new_image = cut_image.reshape(sh[0] // binning_factor, binning_factor, sh[1] // binning_factor, binning_factor).mean(
        -1).mean(1)

    return new_image


def cartesian_to_polar(x, y, center):
    """A function that converts the x,y coordinates to polar ones.
    -Does not do the circular correction
    """
    corrected_x = x-center[0]
    corrected_y = y-center[1]
    theta = np.arctan2(corrected_y, corrected_x)
    r = np.sqrt(corrected_x**2 + corrected_y**2)
    return theta, r


def polar_to_cartesian(r, theta, center):
    """A function that converts polar (r,theta) coordinates to cartesian(x,y) ones

    Parameters
    ----------
    r: float
        radius
    theta: float
        angle
    center: array
        center of the array [x,y]
    Returns
    ----------
    x: float
    y: float
    """
    x = center[0] + r*np.cos(theta)
    y = center[1] + r*np.sin(theta)
    return x, y


def cartesian_list_to_polar(x_list,y_list,center):
    """A function that converts a list of x,y coordinates to (r,theta)

    Parameters
    ----------
    r: float
        radius
    theta: float
        angle
    center: array
        center of the array [x,y]
    Returns
    ----------
    theta_list: 2-d array
    r_list: 2-d array
    """
    theta_list = []
    r_list = []
    for x, y in zip(x_list,y_list):
        t, r = cartesian_to_polar(x,y,center)
        theta_list.append(t)
        r_list.append(r)
    return theta_list, r_list


def polar_list_to_cartesian(r_list, theta_list, center):
    """
    Parameters
    ----------
    r_list: array_like
        radius
    theta_list: array_like
        angle
    center: array
        center of the array [x,y]
    Returns
    ----------
    x_list: array
    y_list: array

    """
    x_list = []
    y_list = []
    for r, t in zip(r_list, theta_list):
        x, y = polar_to_cartesian(r, t, center)
        x_list.append(x)
        y_list.append(y)
    return x_list, y_list


def create_grid(dimension1, dimension2):
    """
    Parameters
    ----------
    dimension1 : array
    dimension2 : array

    Returns
    ----------
    a: array
    b: array
    """
    dim1, dim2 = np.meshgrid(dimension1, dimension2)
    size = len(dimension1)*len(dimension2)
    a, b = np.reshape(dim1, size), np.reshape(dim2,size)
    return a, b


def ellipsoid_list_to_cartesian(r_list, theta_list, center, major, minor, angle, even_spaced=False):
    """Takes a list of ellipsoid points and then use then find their cartesian equivalent

    Parameters
    ----------
    r_list: array
        list of all of the radius.  Can either be all values or even_spaced
    theta_list: array
        list of all of the radius.  Can either be all values or even_spaced
    center: array_like
        center of the ellipsoid
    major: float
        length of the major axis
    minor: float
        length of the minor axis
    angle: float
        angle of the major axis in radians
    even_spaced: bool
        if the values are even spaced.

    Returns
    ----------
    x_list: array_like
        list of x points
    y_list: array_like
        list of y points
    """
    x_list = []
    y_list = []
    focii_ratio = minor / major
    h_o = (1 / ((focii_ratio ** 2) + 1) ** 0.5)  # finding the major and minor axis lengths
    k_o = (1 / ((focii_ratio ** -2) + 1) ** 0.5)
    cos_angle = np.cos(angle)
    sin_angle = np.sin(angle)
    if even_spaced:
        t_sin = [np.sin(t)for t in theta_list]
        t_cos = [np.cos(t)for t in theta_list]
        h_list = np.multiply(r_list, h_o)
        k_list = np.multiply(r_list, k_o)
        x_unrotated = [[h*tc - h*ts for ts, tc in zip(t_sin, t_cos)]for h in h_list]
        y_unrotated = [[k*tc + k*ts for ts, tc in zip(t_sin, t_cos)] for k in k_list]
        x_list = np.subtract(np.multiply(x_unrotated, cos_angle), np.multiply(y_unrotated, sin_angle))
        x_list = np.add(x_list, center[0])
        y_list = np.add(np.multiply(y_unrotated, cos_angle), np.multiply(x_unrotated, sin_angle))
        y_list = np.add(y_list, center[1])
    else:
        for r, t in zip(r_list,theta_list):
            h = h_o *r
            k = k_o *r
            x_unrotated = h*(np.cos(t)) - h*np.sin(t)
            y_unrotated = k*(np.sin(t)) + k*np.cos(t)
            #  need to rotate by the angle
            x = center[0] + x_unrotated * cos_angle - y_unrotated * sin_angle
            y = center[1]+y_unrotated*cos_angle + x_unrotated*sin_angle
            x_list.append(x)
            y_list.append(y)
    return x_list, y_list



import numpy as np


def fem(r_theta_imgs, version="omega", binning=1, cut=40):
    """Calculated the variance among some image

    Parameters
    ----------
    r_theta_imgs: array_like
        polar images
    version: str
        The name of the FEM equation to use
    binning : int
        binning factor
    cut : int
        The cut off in pixels to not consider
    """
    imgs = r_theta_imgs
    imgs = imgs[:, cut // binning:len(imgs), :]
    if version is "omega":
        I_2_avg = np.mean(np.power(imgs, 2), axis=2)
        print(np.shape(I_2_avg))
        I_avg_2 = np.power(np.mean(imgs, axis=2), 2)
        print(np.shape(I_avg_2))
        int_vs_k = np.mean(np.divide(np.subtract(I_2_avg, I_avg_2), I_avg_2), axis=0)
        print(np.shape(int_vs_k))
    else:
        print(np.shape(r_theta_imgs))
        avg2 = np.mean(np.power(np.mean(r_theta_imgs, axis=1), 2), axis=0)
        print(np.shape(avg2))
        avg1 = np.power(np.mean(np. mean(r_theta_imgs, axis=1), axis=0), 2)
        print(np.shape(avg1))
        int_vs_k = np.divide(np.subtract(avg2, avg1), avg1)
        plt.plot(avg2)
        plt.plot(avg1)
        plt.show()
        plt.plot(int_vs_k)
        plt.show()
    return int_vs_k


import hyperspy.api as hs
import numpy as np
import random
from empyer.misc.angular_correlation import angular_correlation

import matplotlib.pyplot as plt


def s_g_kernel(kernel_size, d_hkl, cluster_size, voltage):
    """
    Parameters
    ----------------
    kernel_size: int
        The size of the kernel to create
    d_hkl: float
        The interplanar spacing in n,
    cluster_size: float
        The size of the cluster being calculated.
    """
    wavelength = 1/(6.626*10**-34/np.sqrt(2*9.109*10**-31*voltage*1000*1.602*10**-19*(1+voltage/(2*511)))*10**9)
    print(wavelength)
    ax = np.arange(-kernel_size // 2 + 1., kernel_size // 2 + 1.)
    xx, yy = np.divide(np.meshgrid(ax, ax), kernel_size / 3 * cluster_size)
    scaling = 1/(kernel_size / 3 * cluster_size)
    sg = np.power((np.square(xx) + np.square(yy)), 0.5)
    print(d_hkl)
    #sg_1 = np.power(sg,-1)
    dot = np.subtract(2*wavelength*d_hkl, np.multiply(sg, d_hkl))
    #mag_1 = ((wavelength**2 + d_hkl**2)**.5)
    #mag_2 = np.power(np.add(np.power(np.subtract(wavelength, sg)**2,2), d_hkl**2),.5)
    #mag = np.multiply(mag_1, mag_2)
    #ang = np.divide(dot, mag)
    #d = np.subtract(wavelength, sg)
    #mag = np.power(np.subtract(wavelength**2, np.multiply(cos)),.5)
    angles = np.divide(sg, d_hkl)
    sg_surf = np.multiply(sg, (2 * np.pi * cluster_size))
    kernel = np.power(np.multiply(np.divide((np.sin(sg_surf) -
                                             np.multiply(sg_surf,
                                                         np.cos(sg_surf))),
                                            np.power(sg_surf, 3)), 3), 2)
    dict0 = {'size': kernel_size, 'name': 's_x', 'units': 'nm^-1', 'scale': scaling, 'offset': 0}
    dict1 = {'size': kernel_size, 'name': 's_y', 'units': 'nm^-1', 'scale': scaling, 'offset': 0}
    k = hs.signals.Signal2D(data=kernel, axes=[dict0, dict1])

    return k


def s_g_kern_toAng(kern, d_hkl):
    factor = (d_hkl*np.pi/180)
    dict0 = {'size':  kern.axes_manager[0].size, 'name': 's_x', 'units': 'degrees',
             'scale': kern.axes_manager[0].scale/factor, 'offset': 0}
    dict1 = {'size': kern.axes_manager[0].size, 'name': 's_y', 'units': 'degrees',
             'scale': kern.axes_manager[0].scale/factor,'offset': 0}
    ang = hs.signals.Signal2D(data=kern.data, axes=[dict0,dict1] )
    return ang


def atomic_displacement_kernel(kernel_size, displacement_factor):
    ax = np.arange(-kernel_size // 2 + 1., kernel_size// 2 + 1.)
    xx, yy = np.meshgrid(ax, ax)
    return


# Functions for simulating rotations
def random_rotation():
    u = random.uniform(0, 1)
    v = random.uniform(0, 1)
    p = random.uniform(0, 1)
    alpha = 2 * np.pi * u
    beta = np.arccos(2 * v -1)
    rotation_vector = (np.cos(alpha)*np.sin(beta),np.sin(beta), np.sin(alpha)*np.cos(beta))
    theta =  2 * np.pi * p
    return rotation_vector, theta


def sg(acc_voltage, rotation_vector, theta, k0=(4,0,0)):
    """
    Parameters
    ----------------
    accelerating_voltage: int
        In kV the voltage of the instrument for calculating Ewald's sphere
    theta: float
        The
    phi: float
        The size of the cluster being calculated.
    omega: float
        The size of the cluster being calculated.
    k: float
        The k spacing for the speckle
    r: float
        The radius of the particle
    """
    es_radius = 1/get_wavelength(acc_voltage)

    q1 = np.array([0, k0[0], k0[1], k0[2]])
    q2 = np.array([np.cos(theta/2),
                   rotation_vector[0]*np.sin(theta/2),
                   rotation_vector[1]*np.sin(theta/2),
                   rotation_vector[2]*np.sin(theta/2)])
    q2_conj = np.array([q2[0], -1*q2[1], -1*q2[2], -1*q2[3]])

    q3 = mult_quaternions(mult_quaternions(q2,q1),q2_conj)
    q3 = q3[1:]
    dist = np.sqrt(q3[0]**2+q3[1]**2+(-es_radius - q3[2])**2)
    s = dist-es_radius
    return s


def shape_function(r, s,):
    C = 2*np.pi*s*r
    n = ((3*(np.sin(C)-C*np.cos(C)))/C**3)**2
    return n


def get_wavelength(acc_voltage):
    h = 6.626*10**-34
    m0 = 9.109*10**-31
    e = 1.602*10**-19
    wavelength = h/np.sqrt(2*m0*acc_voltage*1000*e*(1+acc_voltage/(2*511)))*10**9
    return wavelength


def mult_quaternions(Q1,Q2):
    w0,x0,y0,z0 = Q1   # unpack
    w1,x1,y1,z1 = Q2
    return([-x1*x0 - y1*y0 - z1*z0 + w1*w0,
            x1*w0 + y1*z0 - z1*y0 + w1*x0,
            -x1*z0 + y1*w0 + z1*x0 + w1*y0,
            x1*y0 - y1*x0 + z1*w0 +w1*z0])


def simulate_symmetry(symmetry=4, I=1, k=4, r=1, iter = 1000):
    angle = (2*np.pi)/symmetry
    k = [[np.cos(angle*i)*k, np.sin(angle*i)*k, 0] for i in range(symmetry)]
    observed_int = np. zeros(shape=(iter,symmetry))
    for i in range(iter):
        rotation_vector, theta =random_rotation()
        for j, speckle in enumerate(k):
            s = sg(acc_voltage=200, rotation_vector=rotation_vector, theta=theta, k0=speckle)
            observed_int[i, j] = I*shape_function(r=r, s=s)

    return observed_int


from .io import load, to_power_signal, to_correlation_signal, to_polar_signal, to_em_signal, to_diffraction_signal
from .signals.diffraction_signal import DiffractionSignal
from .signals.polar_signal import PolarSignal
from .signals.correlation_signal import CorrelationSignal
from .signals.power_signal import PowerSignal

name = "empyer"



import numpy as np
from empyer.signals.em_signal import EM_Signal
from empyer.signals.correlation_signal import CorrelationSignal
from empyer.misc.angular_correlation import angular_correlation
from empyer.misc.image import square


class PolarSignal(EM_Signal):
    _signal_type = "polar_signal"

    def __init__(self, *args, **kwargs):
        """Create a  Polar Signal from a numpy array.
        Parameters
        ----------
        data : numpy array
           The signal data. It can be an array of any dimensions.
        axes : dictionary (optional)
            Dictionary to define the axes (see the
            documentation of the AxesManager class for more details).
        attributes : dictionary (optional)
            A dictionary whose items are stored as attributes.
        metadata : dictionary (optional)
            A dictionary containing a set of parameters
            that will to stores in the `metadata` attribute.
            Some parameters might be mandatory in some cases.
        original_metadata : dictionary (optional)
            A dictionary containing a set of parameters
            that will to stores in the `original_metadata` attribute. It
            typically contains all the parameters that has been
            imported from the original data file.
        """
        EM_Signal.__init__(self, *args, **kwargs)
        self.metadata.set_item("Signal.type", "polar_signal")

    def autocorrelation(self, binning_factor=1, cut=0, normalize=True):
        # TODO: Add the ability to cutoff like slicing (maybe use np.s)
        """Create a Correlation Signal from a numpy array.

        Parameters
        ----------
        binning_factor : int
            Binning factor to speed up calculations
        cut : int or float
            The number of pixels or distance to cut off image
        normalize : boolean
            normalize with autocorrelation
        Returns
        ----------
        angle : CorrelationSignal

        """
        if isinstance(cut, float):
            cut = self.axes_manager.signal_axes[-3].value2index(cut)
        correlation = self.map(angular_correlation,
                               mask=self.get_mask(),
                               binning=binning_factor,
                               cut_off=cut,
                               normalize=normalize,
                               inplace=False)
        passed_meta_data = self.metadata.as_dictionary()
        if self.metadata.has_item('Mask'):
            del (passed_meta_data['Mask'])
        angular = CorrelationSignal(correlation, metadata=passed_meta_data)
        shift = cut // binning_factor
        angular.axes_manager.navigation_axes = self.axes_manager.navigation_axes
        angular.set_axes(-2,
                         name="Radians",
                         scale=self.axes_manager[-2].scale*binning_factor,
                         units="rad")
        offset = shift * self.axes_manager[-1].scale*binning_factor
        angular.set_axes(-1,
                         name="k",
                         scale=self.axes_manager[-1].scale*binning_factor,
                         units=self.axes_manager[-1].units,
                         offset=offset)
        return angular

    def mask_data(self):
        # TODO: Make a function for getting mean with a described mask without using numpy.
        m = self.get_mask()
        if m is not None:
            self.map(np.ma.masked_array, mask=m, show_progressbar=False)

    def fem(self, version="omega"):
        """Calculated the variance among some image

        Parameters
        ----------
        version : str
            The name of the FEM equation to use. 'rings' calculates the mean of the variances of all the patterns at
            some k.  'omega' calculates the variance of the annular means for every value of k.
        """
        if version is 'rings':
            self.mask_data()
            var = self.nanmean(axis=-1)
            var.map(square)
            var = var.nanmean()
            center = self.nanmean(axis=-1).nanmean()
            center.map(square)
            int_vs_k = (var-center)/center
            print(int_vs_k.axes_manager)
        elif version is 'omega':
            self.mask_data()
            var = self.map(square,show_progressbar=False, inplace=False).nanmean().nanmean(axis=1)
            center = self.nanmean(axis=-1)
            center.map(square)
            center = center.nanmean()
            int_vs_k = (var - center) / center
            print(int_vs_k.axes_manager)
        return int_vs_k




from hyperspy.api import load as hsload
from .signals.em_signal import EM_Signal
from .signals.diffraction_signal import DiffractionSignal
from .signals.polar_signal import PolarSignal
from .signals.correlation_signal import CorrelationSignal
from .signals.power_signal import PowerSignal


def load(filenames=None,
         signal_type=None,
         stack=False,
         stack_axis=None,
         new_axis_name='stack_element',
         lazy=False,
         **kwds):
    """Extends the hyperspy loading functionality with additional ability to load empyer signals.

    Parameters
    ---------------------
    filenames:list
        The list of file paths or single file to be loaded
    signal_type: str
        The type of signal to be loaded. Can be read from file or given
    stack: bool
        Stack creates a new axis and loads a list of files into one signal
    new_axis_name: str
        The name of the new axis created with stack
    lazy: bool
        Load the signal into memory or load chunks.  May effect performance, but useful for large datasets.

    Returns
    ---------------------
    signal:
        A signal of type signal_type
    """
    signal = hsload(filenames=filenames,
                    signal_type=signal_type,
                    stack=stack,
                    stack_axis=stack_axis,
                    new_axis_name=new_axis_name,
                    lazy=lazy,
                    **kwds)
    if signal.metadata.Signal.signal_type == 'diffraction_signal':
        signal = to_diffraction_signal(signal)
    if signal.metadata.Signal.signal_type == 'em_signal':
        signal = to_em_signal(signal)
    if signal.metadata.Signal.signal_type == 'polar_signal':
        signal = to_polar_signal(signal)
    if signal.metadata.Signal.signal_type == 'correlation_signal':
        signal = to_correlation_signal(signal)
    if signal.metadata.Signal.signal_type == 'power_signal':
        signal = to_power_signal(signal)
    if signal.metadata.Signal.has_item('signal_type'):
        print(signal.metadata.Signal.signal_type, " loaded!")
    return signal


def to_em_signal(signal=None):
    """Hyperspy signal to em_signal
    """
    ax = signal.axes_manager.as_dictionary()
    ax = [ax[key]for key in ax]
    ds = EM_Signal(signal, metadata=signal.metadata.as_dictionary(), axes=ax)
    return ds


def to_diffraction_signal(signal=None):
    """Hyperspy signal to diffraction_signal
    """
    ax = signal.axes_manager.as_dictionary()
    ax = [ax[key]for key in ax]
    ds = DiffractionSignal(signal, metadata=signal.metadata.as_dictionary(), axes=ax)
    return ds


def to_polar_signal(signal=None):
    """Hyperspy signal to polar_signal
     """
    ax = signal.axes_manager.as_dictionary()
    ax = [ax[key]for key in ax]
    ps = PolarSignal(signal,
                     metadata=signal.metadata.as_dictionary(),
                     axes=ax)
    return ps


def to_correlation_signal(signal=None):
    """Hyperspy signal to correlation_signal
     """
    ax = signal.axes_manager.as_dictionary()
    ax = [ax[key]for key in ax]
    cs = CorrelationSignal(signal,
                     metadata=signal.metadata.as_dictionary(),
                     axes=ax)
    return cs


def to_power_signal(signal=None):
    """Hyperspy signal to power_signal
     """
    ax = signal.axes_manager.as_dictionary()
    ax = [ax[key]for key in ax]
    ps = PowerSignal(signal,
                     metadata=signal.metadata.as_dictionary(),
                     axes=ax)
    return ps



import numpy as np
import math
import random
import matplotlib.pyplot as plt
from numpy.linalg import eig, inv
from collections import namedtuple
from matplotlib.patches import Ellipse


def find_center(img):
    """
    Function which finds the center of a diffraction pattern

    Using an algorithm described in T.C. Patterson et al. Ultramicroscopy 103, 275 (2005). The algorithm assumes the
    sample is circularly symmetric.Takes a random sampling of positions in a cropped region and compares to other points

    Parameters
    ---------
    img: array-like
        An input image with an imposed circle of higher intensity
    """
    def mean_squared_displacement(iterations, mask_x, mask_y, center, img, no_mask=True):
        count = 0
        msd = 0
        points = len(mask_x)
        for i in range(iterations):
            random_position = random.randint(0, points - 1)
            initial_x, initial_y = mask_x[random_position], mask_y[random_position]
            radius = calculate_magnitude(mask_x[random_position], mask_y[random_position])
            theta = random.uniform(0, 2 * np.pi)
            x_rotated, y_rotated = polar_to_cartesian(theta, radius, center)
            initial_x = int(initial_x + center[0])
            initial_y = int(initial_y + center[1])
            final, initial = img[x_rotated][y_rotated], img[initial_x][initial_y]

            if no_mask or final.mask is True or initial.mask is True:
                msd = msd + (final - initial) ** 2
                count = count + 1

        if count:
            msd = msd / count
        else:
            msd = 0
        return msd

    def create_circular_grid(inner_radius, outer_radius):
        x, y = np.meshgrid(range(-outer_radius, outer_radius + 1), range(-outer_radius, outer_radius + 1))
        r = np.sqrt(x ** 2 + y ** 2)
        inside = r < outer_radius
        x, y = x[inside], y[inside]
        return x, y

    def calculate_magnitude(x, y):
        magnitude = np.sqrt((x) ** 2 + (y) ** 2)
        return magnitude

    def polar_to_cartesian(theta, r, center):
        # could do linear interpolation but way more computationally difficult
        x = int(math.floor(r * np.cos(theta)) + center[0])
        y = int(math.floor(r * np.sin(theta)) + center[1])
        return x, y

    search_radius = 7
    all_msd = []
    shape = np.shape(img)
    image_center = np.divide(shape, 2)
    left_bound, right_bound = int(image_center[0]-(search_radius*4)), int(image_center[1]+(search_radius*4))
    MSD = namedtuple('MSD', 'mean_displacement position')
    for x in range(left_bound, right_bound, 4):
        for y in range(left_bound, right_bound, 4):
            center = [x, y]
            mask_x, mask_y = create_circular_grid(50,150)
            msd = mean_squared_displacement(500, mask_y, mask_y, center, img)
            msd =MSD(mean_displacement=msd, position=[x, y])
            all_msd.append(msd)
    center = min(all_msd, key=lambda k: k.mean_displacement).position
    msd2 = []
    left_bound, right_bound = int(center[0]-10), int(center[1]+10)
    for x in range(left_bound, right_bound, 2):
        for y in range(left_bound, right_bound, 2):
            center = [x, y]
            mask_x, mask_y = create_circular_grid(500, 150)
            msd = mean_squared_displacement(500, mask_y, mask_y, center, img)
            msd =MSD(mean_displacement=msd, position=[x, y])
            msd2.append(msd)
    center = min(msd2, key=lambda k: k.mean_displacement).position
    return center


def solve_ellipse(img, mask=None, interactive=False, num_points=500, plot=False):
    """Takes a 2-d array image and allows you to solve for the equivalent ellipse.

    Fitzgibbon, A. W., Fisher, R. B., Hill, F., & Eh, E. (1999). Direct Least Squres Fitting of Ellipses.
        IEEE Transactions on Pattern Analysis and Machine Intelligence, 21(5), 476–480.
    http://nicky.vanforeest.com/misc/fitEllipse/fitEllipse.html

    Parameters
    ----------
    img : array-like
        Image with ellipse with more intense
    interactive: bool
        Allows you to pick points for the ellipse instead of taking the top 2000 points
    plot: bool
        plots the unwrapped image as well as a super imposed ellipse
    Returns
    ----------
    center: array-like
        In cartesian coordinates or (x,y)!!!!!! arrays are in y,x
    lengths:array-like
        The 'length' in pixels of the major and minor axis
    angle:float
        In radians based on the major axis
    """

    def fit_ellipse(x,y):
        x = x[:, np.newaxis]  # reshaping the x and y axis
        y = y[:, np.newaxis]

        D = np.hstack((x*x, x*y, y*y, x, y, np.ones_like(x)))  # Design matrix [x^2, xy, y^2, x,y,1]
        S = np.dot(D.T, D)  # Scatter Matrix
        C = np.zeros([6,6])
        C[0, 2] = C[2, 0] = 2; C[1, 1] = -1
        E, V = eig(np.dot(inv(S), C)) # eigen decomposition to solve constrained minimization probelm
        n = np.argmax(np.abs(E))  # maximum eigenvalue solution
        a = V[:, n]
        return a

    def ellipse_center(a):
        b, c, d, f, g, a = a[1] / 2, a[2], a[3] / 2, a[4] / 2, a[5], a[0]
        num = b * b - a * c
        x0 = (c * d - b * f) / num
        y0 = (a * f - b * d) / num
        return np.array([x0, y0])

    def ellipse_axis_length(a):
        b, c, d, f, g, a = a[1] / 2, a[2], a[3] / 2, a[4] / 2, a[5], a[0]
        up = 2 * (a * f * f + c * d * d + g * b * b - 2 * b * d * f - a * c * g)
        down1 = (b * b - a * c) * ((c - a) * np.sqrt(1 + 4 * b * b / ((a - c) * (a - c))) - (c + a))
        down2 = (b * b - a * c) * ((a - c) * np.sqrt(1 + 4 * b * b / ((a - c) * (a - c))) - (c + a))
        res1 = np.sqrt(up / down1)
        res2 = np.sqrt(up / down2)
        return np.array([res1, res2])

    def ellipse_angle_of_rotation( a ):
        b,c,d,f,g,a = a[1]/2, a[2], a[3]/2, a[4]/2, a[5], a[0]
        if b == 0:
            if a > c:
                return 0
            else:
                return np.pi/2
        else:
            if a > c:
                return (np.pi/2)+((np.pi/2)-np.arctan((a-c)/(2*b)))/2  # pi/2+0.5*cot^-1((a-c)/2b)
            else:
                return (np.pi/2-np.arctan((a-c)/(2*b)))/2
    img_shape = np.shape(img)
    img_list = np.reshape(img, (-1, *img_shape[-2:]))

    if mask is not None:
        img[mask] = 0
    coords = [[], []]
    if interactive:
        figure1 = plt.figure()
        ax = figure1.add_subplot(111)
        plt.imshow(i)
        plt.text(x=-50, y=-20, s="Click to add points. Click again to remove points. You must have 5 points")

        def add_point(event):
            ix, iy = event.xdata, event.ydata
            x_diff, y_diff = np.abs(np.subtract(coords[0], ix)), np.abs(np.subtract(coords[1], iy))
            truth = list(np.multiply(x_diff < 10, y_diff < 10))
            if not len(x_diff):
                coords[0].append(ix)
                coords[1].append(iy)
            elif not any(truth):
                coords[0].append(ix)
                coords[1].append(iy)
            else:
                truth = list(np.multiply(x_diff < 10, y_diff < 10)).index(True)
                coords[0].pop(truth)
                coords[1].pop(truth)
            print(coords)
            ax.clear()
            ax.imshow(img)
            ax.scatter(coords[0], coords[1], s=10, marker="o", color="crimson")
            figure1.canvas.draw()
        cid = figure1.canvas.mpl_connect('button_press_event', add_point)
        plt.show()
        print(coords[0])
    #  non-interactive, works better if there is an intense ring
    # TODO: Make method more robust with respect to obviously wrong points
    else:
        i_shape = np.shape(img)
        print(i_shape)
        flattened_array = img.flatten()
        indexes = sorted(range(len(flattened_array)), key=flattened_array.__getitem__)
        # take top 5000 points make sure exclude zero beam
        coords[0] = np.remainder(indexes[-num_points:], i_shape[0])  # x axis (column)
        coords[1] = np.floor_divide(indexes[-num_points:], i_shape[1])  # y axis (row)
    a = fit_ellipse(np.array(coords[0]), np.array(coords[1]))
    center = ellipse_center(a)  # (x,y)
    #  center = [center[1],center[0]] # array coordinates (y,x)
    lengths = ellipse_axis_length(a)
    angle = ellipse_angle_of_rotation(a)
    #  angle = (np.pi/2)-angle # transforming to array coordinates
    print("The center is:", center)
    print("The major and minor axis lengths are:", lengths)
    print("The angle of rotation is:", angle)
    if plot:
        ellipse = Ellipse((center[0], center[1]), lengths[0] * 2, lengths[1] * 2, angle=angle, fill=False)
        fig = plt.figure()
        axe = fig.add_subplot(111)
        axe.imshow(img)
        axe.add_patch(ellipse)
        plt.show()
    print(img_shape)
    return center, lengths, angle




from empyer.signals.em_signal import EM_Signal
from empyer.signals.power_signal import PowerSignal
from empyer.misc.angular_correlation import power_spectrum


class CorrelationSignal(EM_Signal):
    """Create a  Correlation Signal from a numpy array.

    Parameters
    ----------
    data : numpy array
       The signal data. It can be an array of any dimensions.
    axes : dictionary (optional)
        Dictionary to define the axes (see the
        documentation of the AxesManager class for more details).
    attributes : dictionary (optional)
        A dictionary whose items are stored as attributes.
    metadata : dictionary (optional)
        A dictionary containing a set of parameters
        that will to stores in the `metadata` attribute.
        Some parameters might be mandatory in some cases.
    original_metadata : dictionary (optional)
        A dictionary containing a set of parameters
        that will to stores in the `original_metadata` attribute. It
        typically contains all the parameters that has been
        imported from the original data file.
    """
    _signal_type = "correlation_signal"

    def __init__(self, *args, **kwargs):
        EM_Signal.__init__(self, *args, **kwargs)
        self.metadata.set_item("Signal.type", "correlation_signal")

    def get_power_spectrum(self, method="FFT"):
        """
        Calculate a power spectrum from the correlation signal

        Parameters
        ----------
        method : str
            'FFT' gives fourier transformation of the angular power spectrum.  Currently the only method available
        """
        power_signal = self.map(power_spectrum,
                                method=method,
                                inplace=False,
                                show_progressbar=False)
        passed_meta_data = self.metadata.as_dictionary()
        if self.metadata.has_item('Masks'):
            del (passed_meta_data['Masks'])
        power = PowerSignal(power_signal)
        power.axes_manager.navigation_axes = self.axes_manager.navigation_axes

        power.set_axes(-2,
                       name="FourierCoefficient",
                       scale=1,
                       units="a.u.",
                       offset=.5)
        power.set_axes(-1,
                       name="k",
                       scale=self.axes_manager[-1].scale,
                       units=self.axes_manager[-1].units,
                       offset=self.axes_manager[-1].offset)
        return power

    def get_summed_power_spectrum(self):
        # TODO: Add in the ability to get the summed power spectrum over an axis.
        summed_pow = self.sum().get_power_spectrum()
        return summed_pow



import numpy as np
import matplotlib.pyplot as plt

from hyperspy.signals import Signal2D


class EM_Signal(Signal2D):
    """
    The Diffraction Signal class extends the Hyperspy 2d signal class
    """
    _signal_type = "em_signal"

    def __init__(self, *args, **kwargs):
        """Basic unit of data in the program.

        Spectrums can be any dimension of array but all of the data should be of the
        same type from the same area... (Maybe define this better later... You can use it kind of however you want though)
        Extends the Signal2D Class from hyperspy so there is that added functionality

        Parameters
        ----------
        data : numpy array
           The signal data. It can be an array of any dimensions.
        axes : dictionary (optional)
            Dictionary to define the axes (see the
            documentation of the AxesManager class for more details).
        attributes : dictionary (optional)
            A dictionary whose items are stored as attributes.
        metadata : dictionary (optional)
            A dictionary containing a set of parameters
            that will to stores in the `metadata` attribute.
            Some parameters might be mandatory in some cases.
        original_metadata : dictionary (optional)
            A dictionary containing a set of parameters
            that will to stores in the `original_metadata` attribute. It
            typically contains all the parameters that has been
            imported from the original data file.
        Notes: For more parameters see hyperspy's Signal2D Class
        """
        Signal2D.__init__(self, *args, **kwargs)

    def set_axes(self, index, name=None, scale=None, units=None, offset=None):
        """Set axes of the signal

        Parameters
        ----------
        index: int
           The index of the axes
        name: str
            The name of the axis
        scale : float
            The scale fo the axis
        units : str
            The units of the axis
        offset : float
            The offset of the axes
        """
        if name is not None:
            self.axes_manager[index].name = name
        if scale is not None:
            self.axes_manager[index].scale = scale
        if units is not None:
            self.axes_manager[index].units = units
        if offset is not None:
            self.axes_manager[index].offset = offset

    def set_mask(self, mask):
        """Set a mask

        Parameters
        ----------
        mask: n x m boolean array
            A boolean array which is set as the mask for some signal.  Will overwrite the mask. Function accepts m x n
            and n x m arrays and will take the transpose of the signal so that it matches the hyperspy signal axes.
        """
        if not self.metadata.has_item('Mask'):
            self.metadata.add_node('Mask')
        if tuple(reversed(self.axes_manager.signal_shape)) == np.shape(mask):
            self.metadata.Mask = mask
        elif self.axes_manager.signal_shape == np.shape(mask):
            self.metadata.Mask = np.transpose(mask)
        else:
            print('The mask shape must be the same shape as the signal')

    def show_mask(self):
        """Plots the mask in a separate window
        """
        # TODO:Show mask overlaid over the signal
        ax = self.axes_manager.signal_axes
        ax = [a.get_axis_dictionary() for a in reversed(ax)]
        Signal2D(data=self.get_mask(), axes=ax).plot()
        plt.show()
        return

    def mask_below(self, value, unmask=False):
        """Applies a mask to every pixel with an average value below value

        Parameters
        ----------
        value: float
            The maximum pixel value to apply a mask to.
        unmask: bool
            Unmask any pixel with a value below value
        """
        if not self.metadata.has_item('Mask'):
            self.metadata.add_node('Mask')
            mask = np.zeros(shape=tuple(reversed(self.axes_manager.signal_shape)), dtype=bool)
            self.metadata.Mask = mask
        if unmask is False:
            is_below = self.mean().data < value
            self.metadata.Mask[is_below] = True
        if unmask is True:
            is_below = self.mean().data < value
            self.metadata.Mask[is_below] = False

    def mask_shape(self, shape='rectangle', data=[1, 1, 1, 1], unmask=False):
        # TODO: Add more shapes
        """Applies a mask to every pixel using a shape and the appropriate definition


        Parameters
        ----------
        shape: str
            Acceptable shapes ['rectangle, 'circle']
        data: list
            Define shapes. eg 'rectangle' -> [x1,x2,y1,y2] 'circle' -> [radius, x,y]
            data allows indexing with floats and the axes described for the signal
        unmask: bool
            Unmask any pixels in the defined shape
        """
        if shape is 'rectangle':
            self.mask_slice(data[0], data[1], data[2], data[3], unmask=unmask)
            return
        if not self.metadata.has_item('Mask'):
            self.metadata.add_node('Mask')
            mask = np.zeros(shape=tuple(reversed(self.axes_manager.signal_shape)), dtype=bool)
            self.metadata.Mask = mask
        if shape is 'circle':
            if not all(isinstance(item, int) for item in data):
                radius = self.axes_manager.signal_axes[0].value2index(data[0])
                y = self.axes_manager.signal_axes[0].value2index(data[1])
                x = self.axes_manager.signal_axes[1].value2index(data[2])
            else:
                radius = data[0]
                y = data[1]
                x = data[2]
            x_ind, y_ind = np.meshgrid(range(-radius, radius + 1), range(-radius, radius + 1))
            r = np.sqrt(x_ind ** 2 + y_ind ** 2)
            inside = r < radius
            x_ind, y_ind = x_ind[inside]+int(x), y_ind[inside]+int(y)
            self.metadata.Mask[x_ind, y_ind] = True
        return

    def mask_slice(self, x1, x2, y1, y2, unmask=False):
        """Applies a mask to some slice of the data (same as mask_shape for shape= 'rectangle')

        Parameters
        ----------
        x1: float or int
        x2: float or int
        y1: float or int
        y2: float or int
        unmask: bool
            Unmask any pixels in the defined shape
        """
        if not self.metadata.has_item('Mask'):
            self.metadata.add_node('Mask')

            mask = np.zeros(shape=tuple(reversed(self.axes_manager.signal_shape)), dtype=bool)
            self.metadata.Mask = mask
        if not all(isinstance(item, int) for item in [x1, x2, y1, y2]):
            x1 = self.axes_manager.signal_axes[0].value2index(x1)
            x2 = self.axes_manager.signal_axes[0].value2index(x2)
            y1 = self.axes_manager.signal_axes[1].value2index(y1)
            y2 = self.axes_manager.signal_axes[1].value2index(y2)
        if unmask is False:
            self.metadata.Mask[y1:y2, x1:x2] = True
        if unmask is True:
            self.metadata.Mask[y1:y2, x1:x2] = False

    def get_signal_axes_values(self):
        """ Returns the values for each pixel of the signal.  Useful for plotting without using hyperspy

        Returns
        ----------
        axis0: array-like
            The values for axis 0
        axis1: array-like
            The values for axis 1
        """
        axis0 = np.linspace(start=self.axes_manager.signal_axes[0].offset,
                            stop=(self.axes_manager.signal_axes[0].size *
                                  self.axes_manager.signal_axes[1].scale +
                                  self.axes_manager.signal_axes[0].offset),
                            num=self.axes_manager.signal_axes[0].size)
        axis1 = np.linspace(start=self.axes_manager.signal_axes[1].offset,
                            stop=(self.axes_manager.signal_axes[1].size *
                                  self.axes_manager.signal_axes[1].scale +
                                  self.axes_manager.signal_axes[1].offset),
                            num=self.axes_manager.signal_axes[1].size)
        return axis0, axis1

    def get_mask(self):
        """ Returns the defined mask for the signal or None if no mask is defined

              Returns
              ----------
              mask: array-like
              """
        if self.metadata.has_item('Mask'):
            mask = self.metadata.Mask
        else:
            return None
        return mask



import numpy as np

from empyer.signals.em_signal import EM_Signal


class PowerSignal(EM_Signal):
    _signal_type = "power_signal"

    def __init__(self, *args, **kwargs):
        # TODO: Add in particle analysis to maps
        """Create a  Power Signal from a numpy array.
        Parameters
        ----------
        data : numpy array
           The signal data. It can be an array of any dimensions.
        axes : dictionary (optional)
            Dictionary to define the axes (see the
            documentation of the AxesManager class for more details).
        attributes : dictionary (optional)
            A dictionary whose items are stored as attributes.
        metadata : dictionary (optional)
            A dictionary containing a set of parameters
            that will to stores in the `metadata` attribute.
            Some parameters might be mandatory in some cases.
        original_metadata : dictionary (optional)
            A dictionary containing a set of parameters
            that will to stores in the `original_metadata` attribute. It
            typically contains all the parameters that has been
            imported from the original data file.
        """
        EM_Signal.__init__(self, *args, **kwargs)
        self.metadata.set_item("Signal.type", "power_signal")

    def get_i_vs_k(self, symmetry=None):
        """ Get the intensity versus k for the summed diffraction patterns

        Parameters
        ----------
        symmetry: int or array-like
            specific integers or list of symmetries to average over when creating the map of the correlations.
        Returns
        ----------
        i: Signal-2D
            The intensity as a function of k for some signal
        """
        if symmetry is None:
            i = self.isig[:, :].sum(axis=[0, 1, 2])
        else:
            i = self.isig[:, symmetry].sum(axis=[0, 1, 2])

        return i

    def get_map(self, k_region=[3.0, 6.0], symmetry=None):
        # TODO: symmetry is still broken :/ I don't know why it doesn't like using arrays to slice....
        """Creates a 2 dimensional map of from the power spectrum.

        Parameters
        ----------
        k_region: array-like
           upper and lower k values to integrate over, allows both ints and floats for indexing
        symmetry: int or array-like
            specific integers or list of symmetries to average over when creating the map of the correlations.
        Returns
        ----------
        symmetry_map: 2-d array
            2 dimensional map of from the power spectrum
        """
        if symmetry is None:

            sym_map = self.isig[k_region[0]:k_region[1], :].sum(axis=[-1,-2]).transpose()
        else:
            print(self.isig[k_region[0]:k_region[1],:].isig[:,[5,6]])
            sym_map = self.isig[k_region[0]:k_region[1], symmetry].sum(axis=[-1, -2]).transpose()
        return sym_map



from setuptools import setup


def readme():
    with open('README.rst') as f:
        return f.read()


setup(name='empyer',
      version='0.163',
      description='Electron Microscopy tools for analyzing 4 and 5 dimensional STEM datasets',
      long_description=readme(),
      keywords='STEM Electron Microscopy Glass',
      url='https://github.com/CSSFrancis/empyer',
      author='CSSFrancis',
      author_email='csfrancis@wisc.edu',
      liscense='MIT',
      packages=['empyer',
                'empyer.misc',
                'empyer.signals',
                'empyer.simulate',
                'empyer.tests'],
      install_requires=['hyperspy >=1.3',
                        'numpy',
                        'matplotlib'],
      include_package_data=True,
      zip_safe=False)



import numpy as np

from empyer.misc.ellipse_analysis import solve_ellipse
from empyer.misc.cartesain_to_polar import convert
from empyer.signals.em_signal import EM_Signal
from empyer.signals.polar_signal import PolarSignal


class DiffractionSignal(EM_Signal):
    """
    The Diffraction Signal class extends the Hyperspy 2d signal class
    """
    _signal_type = "diffraction_signal"

    def __init__(self, *args, **kwargs):
        """Create a Diffraction Signal from a numpy array.

        Parameters
        -------
        data : numpy array
           The signal data. It can be an array of any dimensions.
        axes : dictionary, optional
            Dictionary to define the axes (see the
            documentation of the AxesManager class for more details).
        attributes : dictionary, optional
            A dictionary whose items are stored as attributes.
        metadata : dictionary, optional
            A dictionary containing a set of parameters
            that will to stores in the `metadata` attribute.
            Some parameters might be mandatory in some cases.
        original_metadata : dictionary, optional
            A dictionary containing a set of parameters
            that will to stores in the `original_metadata` attribute. It
            typically contains all the parameters that has been
            imported from the original data file.
        """
        EM_Signal.__init__(self, *args, **kwargs)
        if not hasattr(self.metadata.Signal, 'Ellipticity.calibrated'):
            self.metadata.set_item("Signal.Ellipticity.calibrated", False)

        if not hasattr(self.metadata.Signal, 'type'):
            self.metadata.set_item("Signal.type", "diffraction_signal")

    def determine_ellipse(self, num_points=500, interactive=False, plot=False):
        # TODO: Identify if there needs to be a drift correction applied.
        # TODO: Exclude the zero beam if it isn't masked.
        """Determine the elliptical nature of the diffraction pattern.

        Parameters
        -------
        interactive : Boolean
            'interactive' nature means that points are chosen to create a ring
        axis : int
            'axis' to determine ellipse along
        num_points : int
            number of points to define ellipse by (only used if interactive = False)
        plot : Boolean
            Weather or not to plot the ellipse

        Returns
        -------
        center : list of int
            the center of the ellipse
        lengths : list of int
            the length in pixels of the major and minor axes
        angle : float
            the angle of the major axes
        """
        center, lengths, angle = solve_ellipse(self.sum().data,
                                               mask=self.get_mask(),
                                               num_points=num_points,
                                               interactive=interactive,
                                               plot=plot)
        self.metadata.set_item("Signal.Ellipticity.center", center)
        self.metadata.set_item("Signal.Ellipticity.angle", angle)
        self.metadata.set_item("Signal.Ellipticity.lengths", lengths)
        self.metadata.set_item("Signal.Ellipticity.calibrated", True)
        return center, lengths, angle

    def calculate_polar_spectrum(self, phase_width=720, radius=None, parallel=False, inplace=False):
        """Take the Diffraction Pattern and unwrap the diffraction pattern.

        Parameters
        -------
        phase_width: int
            The number of pixels in the x direction
        radius: int
            The number of pixels in the y direction
        parallel: boolean
            use multiple processors for calculations (useful for large numbers of diffraction patterns,
            more for large pixel size)
        inplace: boolean
            replaces diffraction pattern data with polar equivalent

        Returns
        -------
        polar: PolarSignal
            Polar signal returned
        """

        if not self.metadata.Signal.Ellipticity.calibrated:
            self.determine_ellipse()
        polar_signal = self.map(convert,
                                mask=self.get_mask(),
                                center=self.metadata.Signal.Ellipticity.center,
                                angle=self.metadata.Signal.Ellipticity.angle,
                                foci=self.metadata.Signal.Ellipticity.lengths,
                                phase_width=phase_width,
                                radius=radius,
                                parallel=parallel,
                                inplace=inplace,
                                show_progressbar=False)
        if self.metadata.has_item('Mask'):
            new_mask = convert(self.get_mask(),
                               center=self.metadata.Signal.Ellipticity.center,
                               angle=self.metadata.Signal.Ellipticity.angle,
                               foci=self.metadata.Signal.Ellipticity.lengths,
                               phase_width=phase_width,
                               radius=radius)
            new_mask = new_mask > 0
            new_mask = np.transpose(new_mask)
        else:
            new_mask = None

        passed_meta_data = self.metadata.as_dictionary()
        if self.metadata.Signal.has_item('Ellipticity'):
            del(passed_meta_data['Signal']['Ellipticity'])
        if self.metadata.has_item('Mask'):
            del (passed_meta_data['Mask'])

        polar = PolarSignal(polar_signal, metadata=passed_meta_data)
        polar.set_mask(mask=new_mask)

        polar.axes_manager.navigation_axes = self.axes_manager.navigation_axes
        polar.set_axes(-2,
                       name="Radians",
                       scale=2*np.pi/phase_width,
                       units="rad")
        polar.set_axes(-1,
                       name="k",
                       scale=self.axes_manager[-1].scale,
                       units=self.axes_manager[-1].units)
        return polar
